#include "MeshViewer.h"
#include "MeshViewerMenu.h"

#include <MESH_Controller.h>
#include <MESH_View.h>
#include <X3DTK/kernel.h>
#include <qgl.h>
#include <qwidget.h>
#include <qmenubar.h>
#include <qlabel.h>
#include <qcursor.h>
#include <qpopupmenu.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qpixmap.h>
#include <qpainter.h>

#include <iostream>

using namespace std;

MeshViewer::MeshViewer(QWidget *parent, const char *name, const QGLWidget *shareWidget, WFlags flags)
: QGLViewer(parent, name, shareWidget, flags), _controller(0), _contextMenu(0)
{
  setMouseBinding(Qt::ControlButton | Qt::LeftButton, CAMERA, TRANSLATE);
}

MeshViewer::~MeshViewer()
{
  for (std::list<X3DTK::MESH::View *>::iterator it = _viewList.begin(); it != _viewList.end(); ++it)
    RC::Signal::disconnect(*it, &X3DTK::MESH::View::updated, this, &QGLWidget::updateGL);      
}

void MeshViewer::init()
{
  glPolygonOffset(-2.0, -2.0);
#ifdef GL_RESCALE_NORMAL
  glEnable(GL_RESCALE_NORMAL);
#endif
  glDisable(GL_LIGHTING);
}

void MeshViewer::setController(X3DTK::MESH::Controller *controller)
{
  _controller = controller;
}

void MeshViewer::record(X3DTK::MESH::View *view)
{
  _viewList.push_back(view);
  if (_controller != 0)
    _controller->record(view);
  
  RC::Signal::connect(view, &X3DTK::MESH::View::updated, this, &QGLWidget::updateGL);
}

void MeshViewer::remove(X3DTK::MESH::View *view)
{
  list<X3DTK::MESH::View *>::iterator it = std::find(_viewList.begin(), _viewList.end(), view);
  if (it != _viewList.end())
    _viewList.erase(it);
  
  RC::Signal::disconnect(view, &X3DTK::MESH::View::updated, this, &QGLWidget::updateGL);  
}

void MeshViewer::draw()
{
  for (list<X3DTK::MESH::View *>::iterator it = _viewList.begin(); it != _viewList.end(); ++it)
    if ((*it)->isVisible())
      (*it)->display();
}

void MeshViewer::select(const QMouseEvent *e)
{
  // Make openGL context current
  makeCurrent();

  const int SENSITIVITY = 5;

  // Loads the matrices
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);
  camera()->getViewport(viewport);
  gluPickMatrix(static_cast<GLdouble>(e->x()), static_cast<GLdouble>(e->y()), SENSITIVITY, SENSITIVITY, viewport);

  // Don't use loadProjectionMatrix() directly as it clears the GL_PROJECTION matrix with a glLoadIdentity.
  // The false flag indicates that no glLoadIdentity should be called, in order to combine the matrices.
  camera()->loadProjectionMatrix(false);
  camera()->loadModelViewMatrix();

  // calling the select
  for (list<X3DTK::MESH::View *>::iterator it = _viewList.begin(); it != _viewList.end(); ++it)
  {
    if ((*it)->isVisible())
      (*it)->select();
  }    
  _controller->select();
      
  // calling the initSelect    
  for (list<X3DTK::MESH::View *>::iterator it = _viewList.begin(); it != _viewList.end(); ++it)
  {
    if ((*it)->isVisible())
      (*it)->initSelect();    
  }
}

void MeshViewer::setItem(int i)
{
  if (_viewMap[i]->isVisible())
    _viewMap[i]->hide();
  else
    _viewMap[i]->show();  
}

void MeshViewer::mouseDoubleClickEvent(QMouseEvent *e)
{
  if (e->button() == Qt::LeftButton)
    RC::Signal::send(this, &MeshViewer::doubleClicked);
  else
    QGLViewer::mouseDoubleClickEvent(e); 
}

void MeshViewer::contextMenuEvent(QContextMenuEvent *)
{
  _contextMenu = new MeshViewerMenu(this);
  _contextMenu->setCheckable(TRUE);
  QLabel *caption = new QLabel("views", this);
  caption->setAlignment(Qt::AlignCenter);
  _contextMenu->insertItem(caption);
  _contextMenu->insertSeparator();
  
  for (list<X3DTK::MESH::View *>::iterator it = _viewList.begin(); it != _viewList.end(); ++it)
  {
    int i = _contextMenu->insertItem(static_cast<const char *>((*it)->getName()));
    _viewMap[i] = *it;
    
    if ((*it)->isVisible())
      _contextMenu->setItemChecked(i, TRUE);
    else
      _contextMenu->setItemChecked(i, FALSE);  
  }  

  RC::Signal::connect(_contextMenu, &MeshViewerMenu::getItem, this, &MeshViewer::setItem);
    
  _contextMenu->exec(QCursor::pos());
  _contextMenu->clear();
  _viewMap.clear();
  
  RC::Signal::disconnect(_contextMenu, &MeshViewerMenu::getItem, this, &MeshViewer::setItem);
  delete _contextMenu;
}
