namespace X3DTK {
namespace MESH {

template<class MData, class VData, class EData, class FData, bool RW>
FacesView<MData, VData, EData, FData, RW>::FacesView(QMainWindow *parent) 
: View("faces"), dock(new MeshFacesDock(parent)), _modified(true), _modifiedSelection(true), _displayList(0), _selectionDisplayList(0), _selectedMesh(0), _selectedFace(0), _selectedTransform(SFMatrix34f::identity)
{
  _facesDrawer = Singleton<TemplateFacesDrawer<MData, VData, EData, FData, RW> >::getInstance();

  RC::Signal::connect(dock, &MeshFacesDock::real, this, &FacesView<MData, VData, EData, FData, RW>::real);
  RC::Signal::connect(dock, &MeshFacesDock::indexedFaceSet, this, &FacesView<MData, VData, EData, FData, RW>::indexedFaceSet);
  RC::Signal::connect(dock, &MeshFacesDock::connComs, this, &FacesView<MData, VData, EData, FData, RW>::connComs);
}

template<class MData, class VData, class EData, class FData, bool RW>
FacesView<MData, VData, EData, FData, RW>::~FacesView()
{
  RC::Signal::disconnect(dock, &MeshFacesDock::real, this, &FacesView<MData, VData, EData, FData, RW>::real);
  RC::Signal::disconnect(dock, &MeshFacesDock::indexedFaceSet, this, &FacesView<MData, VData, EData, FData, RW>::indexedFaceSet);
  RC::Signal::disconnect(dock, &MeshFacesDock::connComs, this, &FacesView<MData, VData, EData, FData, RW>::connComs);

  glDeleteLists(_displayList, 1);
  glDeleteLists(_selectionDisplayList, 1);
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::receiveNotification()
{
  _modified = true;
  _modifiedSelection = true;
  glDeleteLists(_displayList, 1);
  glDeleteLists(_selectionDisplayList, 1);
  _displayList = glGenLists(1);
  _selectionDisplayList = glGenLists(1);
  _selectedMesh = 0;
  _selectedFace = 0;
  _selectedTransform = SFMatrix34f::identity;
  _facesDrawer->changeModel();
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::display()
{
  if (_modified)
  {
    glNewList(_displayList, GL_COMPILE_AND_EXECUTE);
    _facesDrawer->draw(getScene());
    glEndList();
    _modified = false;
  }  
  else
    glCallList(_displayList);
  
  _facesDrawer->drawSelected(getScene());
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::initSelect()
{
  _selectedMesh = 0;
  _selectedFace = 0;
  _selectedTransform = SFMatrix34f::identity;
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::select()
{
  const int NB_HITS_MAX = 32768;
  // Prepare the selection mode
  static GLuint hits[NB_HITS_MAX];

  glSelectBuffer(NB_HITS_MAX, hits);
  glRenderMode(GL_SELECT);
  glInitNames();

  if (_modifiedSelection)
  {
    glNewList(_selectionDisplayList, GL_COMPILE_AND_EXECUTE);
    _facesDrawer->draw(getScene(), true);
    glEndList();
    _modifiedSelection = false;
  }  
  else
    glCallList(_selectionDisplayList);

  glFlush();

  // Get the results
  GLint nb_hits = glRenderMode(GL_RENDER);

  // Too many elements in selection buffer
  if (nb_hits < 0)
    return;

  // No selection
  if (nb_hits == 0)
  {
    _selectedMesh = 0;
    _selectedFace = 0;
    _selectedTransform = SFMatrix34f::identity;
    return;
  }
 
  //interpreting the results.
  _zMin = hits[2];  
  int selectedMesh(hits[3]);
  int selectedId(hits[4]);

  for (int i = 1; i < nb_hits; ++i)
  {
    // Prefer vertices over edges and edges over faces
    if (hits[i*5 + 1] < _zMin)
    {
      _zMin = hits[i*5 + 2];
      selectedMesh = hits[i*5 + 3];
      selectedId = hits[i*5 + 4];
    }
  }  

  _selectedMesh = _facesDrawer->getMesh(selectedMesh);
  _selectedFace = _selectedMesh->getFaces()[selectedId];
  _selectedTransform = _facesDrawer->getMatrix(selectedMesh);
}

template<class MData, class VData, class EData, class FData, bool RW>
unsigned int FacesView<MData, VData, EData, FData, RW>::getValue() const
{
  if (_selectedFace != 0)
    return _zMin;
  return std::numeric_limits<unsigned int>::max();  
}

template<class MData, class VData, class EData, class FData, bool RW>
SFTemplateFace<MData, VData, EData, FData, RW> *FacesView<MData, VData, EData, FData, RW>::getSelectedFace() const 
{
  return _selectedFace;
}

template<class MData, class VData, class EData, class FData, bool RW>
TemplateMesh<MData, VData, EData, FData, RW> *FacesView<MData, VData, EData, FData, RW>::getSelectedMesh() const 
{
  return _selectedMesh;
}

template<class MData, class VData, class EData, class FData, bool RW>
const SFMatrix34f &FacesView<MData, VData, EData, FData, RW>::getSelectedTransform() const 
{
  return _selectedTransform;
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::setSelectedNodes(const MFNode &selectedNodes)
{
  _facesDrawer->setSelectedNodes(selectedNodes);
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::real()
{
  _facesDrawer->setRenderingMode(REAL);
  notify();
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::indexedFaceSet()
{
  _facesDrawer->setRenderingMode(IFS);
  notify();  
}

template<class MData, class VData, class EData, class FData, bool RW>
void FacesView<MData, VData, EData, FData, RW>::connComs()
{
  _facesDrawer->setRenderingMode(CONNCOMS);
  notify();
}
  
}
}
